/* ---------------------------------------------------------------------
BMP2RAG.C (C) Copyright Bill Buckels 2013
All Rights Reserved.

Licence Agreement
-----------------

You have a royalty-free right to use, modify, reproduce and
distribute this source code in any way you find useful,
provided that you agree that Bill Buckels has no warranty obligations
or liability resulting from said distribution in any way whatsoever.
If you don't agree, remove this source code and related files
from your computer now.

Written by   : Bill Buckels
Email:         bbuckels@escape.ca

Purpose      : One way utility to convert Windows 16 Color BMP files
			   To Apple II HGR files.


This utility produces full-screen HGR files for Graphics-Only and Mixed
Text and Graphics in 4 different formats from 16 color, 256 color, and
24-bit color Windows BMP files.

It also produces partial screen image fragments in 3 different formats
(sorry AppleSoft Programmers but you can optionally output a BSaved full
screen with the fragment in the top-left corner).

Quite a number of features are available and of particular interest for
the C programmer is the optional embedded graphics output... which can be
inserted directly into a Apple II C program as a byte array.

16 color test BMP files are provided in the 3 - 16 color base resolutions.

The source code is well-commented and should prove informative as well.

Bill Buckels
February 18,2013

Revision     : 1.0 First Release
-----------------------------------------------------------------------
Written in Large Model 16 bit Microsoft C (MSC) Version 8.00c
Note: Run in an MS-DOS emulator like DOSBox if you can't run it raw.
----------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int BUGS=0;

#define SUCCESS  0
#define VALID    SUCCESS
#define FAILURE  -1
#define INVALID  FAILURE
#define ASCIIZ 0

/* Bitmap Header structures */
typedef struct tagRGBQUAD
{
    unsigned char    rgbBlue;
    unsigned char    rgbGreen;
    unsigned char    rgbRed;
    unsigned char    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER
{
    unsigned long   biSize;
    unsigned long   biWidth;
    unsigned long   biHeight;
    unsigned int    biPlanes;
    unsigned int    biBitCount;
    unsigned long   biCompression;
    unsigned long   biSizeImage;
    unsigned long   biXPelsPerMeter;
    unsigned long   biYPelsPerMeter;
    unsigned long   biClrUsed;
    unsigned long   biClrImportant;
} BITMAPINFOHEADER;

/* constants for the biCompression field

   Unlikely to run into any RLE BMP's and it doesn't matter because I'm
   not supporting them anyway because I'm lazy to some degree. */
#define BI_RGB      0L
#define BI_RLE8     1L
#define BI_RLE4     2L

typedef struct tagVGABITMAPINFO
{
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD          bmiColors[16];
} VGABITMAPINFO;


typedef struct tagBITMAPFILEHEADER
{
    unsigned char   bfType[2];
    unsigned long   bfSize;
    unsigned int    bfReserved1;
    unsigned int    bfReserved2;
    unsigned long   bfOffBits;
} BITMAPFILEHEADER;

/* these constants are what I use in this program but have a limited
   meaning in reality and therefore should not be taken as anything but
   process control ordinals... they do not hold the mystery of the
   universe and are simply an ordering of the 6 Apple II HGR colors */

enum hgrcolors{
	HBLACK=0,
	HGREEN,
	HVIOLET,
	HORANGE,
	HBLUE,
	HWHITE};

unsigned char *szTextTitle =
    "BMP2RAG(C) Version 1.0 Copyright Bill Buckels 2013. All Rights Reserved.";

unsigned char *szUsage =
    	"BMP2RAG produces Apple II HGR files from Windows BMP files.\n"
		"Usage is \"BMP2RAG MY.BMP [options]\"\n"
    	"Please read the BMP2RAG documentation and source code for\n"
      	"additional information about command line options and features.\n";

/* static structures for processing input file */
BITMAPFILEHEADER BitMapFileHeader;
RGBQUAD          sbmp[256]; /* super vga - applewin */
VGABITMAPINFO    bmp;
BITMAPINFOHEADER bmi;

/* globals for image processing */
unsigned maxwidth = 280, maxheight = 192, bmpformat = 280, ragwidth, ragheight, ragpixels,
		 textoutput = 0, binoutput = 0, x14 = 0,
		 externalmap = 0, applewindex =0, applewinversion = 0, a2shift=0,
		 Breverse=0, Vreverse=0, Wreverse=0, doublecolors=0,doubleblack=0;


/* translation table for remapping of bmp scanlines to
   apple II rasters */
unsigned char bmpmapinfo[16] = {255,255,255,255,255,255,255,255,
                                255,255,255,255,255,255,255,255};


/* provides base address for page1 hires scanlines  */
unsigned HB[]={
0x2000, 0x2400, 0x2800, 0x2C00, 0x3000, 0x3400, 0x3800, 0x3C00,
0x2080, 0x2480, 0x2880, 0x2C80, 0x3080, 0x3480, 0x3880, 0x3C80,
0x2100, 0x2500, 0x2900, 0x2D00, 0x3100, 0x3500, 0x3900, 0x3D00,
0x2180, 0x2580, 0x2980, 0x2D80, 0x3180, 0x3580, 0x3980, 0x3D80,
0x2200, 0x2600, 0x2A00, 0x2E00, 0x3200, 0x3600, 0x3A00, 0x3E00,
0x2280, 0x2680, 0x2A80, 0x2E80, 0x3280, 0x3680, 0x3A80, 0x3E80,
0x2300, 0x2700, 0x2B00, 0x2F00, 0x3300, 0x3700, 0x3B00, 0x3F00,
0x2380, 0x2780, 0x2B80, 0x2F80, 0x3380, 0x3780, 0x3B80, 0x3F80,
0x2028, 0x2428, 0x2828, 0x2C28, 0x3028, 0x3428, 0x3828, 0x3C28,
0x20A8, 0x24A8, 0x28A8, 0x2CA8, 0x30A8, 0x34A8, 0x38A8, 0x3CA8,
0x2128, 0x2528, 0x2928, 0x2D28, 0x3128, 0x3528, 0x3928, 0x3D28,
0x21A8, 0x25A8, 0x29A8, 0x2DA8, 0x31A8, 0x35A8, 0x39A8, 0x3DA8,
0x2228, 0x2628, 0x2A28, 0x2E28, 0x3228, 0x3628, 0x3A28, 0x3E28,
0x22A8, 0x26A8, 0x2AA8, 0x2EA8, 0x32A8, 0x36A8, 0x3AA8, 0x3EA8,
0x2328, 0x2728, 0x2B28, 0x2F28, 0x3328, 0x3728, 0x3B28, 0x3F28,
0x23A8, 0x27A8, 0x2BA8, 0x2FA8, 0x33A8, 0x37A8, 0x3BA8, 0x3FA8,
0x2050, 0x2450, 0x2850, 0x2C50, 0x3050, 0x3450, 0x3850, 0x3C50,
0x20D0, 0x24D0, 0x28D0, 0x2CD0, 0x30D0, 0x34D0, 0x38D0, 0x3CD0,
0x2150, 0x2550, 0x2950, 0x2D50, 0x3150, 0x3550, 0x3950, 0x3D50,
0x21D0, 0x25D0, 0x29D0, 0x2DD0, 0x31D0, 0x35D0, 0x39D0, 0x3DD0,
0x2250, 0x2650, 0x2A50, 0x2E50, 0x3250, 0x3650, 0x3A50, 0x3E50,
0x22D0, 0x26D0, 0x2AD0, 0x2ED0, 0x32D0, 0x36D0, 0x3AD0, 0x3ED0,
0x2350, 0x2750, 0x2B50, 0x2F50, 0x3350, 0x3750, 0x3B50, 0x3F50,
0x23D0, 0x27D0, 0x2BD0, 0x2FD0, 0x33D0, 0x37D0, 0x3BD0, 0x3FD0};

/* hgrbuf is essentially a copy of the Apple II HGR screen memory complete
   with screen holes... it is basically an Apple II BSaved Image waiting
   to be written... after BMP2RAG builds it of course. */
unsigned char hgrbuf[8192];

unsigned char bmpscanline[140]; /* up to 140 bytes will be read */

/* the following are used for double-buffering of translated BMP pixels
   during processing... after emptying (and scaling) buf280 into work280
   buf280 is used again to accumulate the values that drive my logic for
   bitwise processing and the final pass that packs the pixels into hgrbuf
   in Apple II hgr formatted rasters */

unsigned char buf280[564], work280[564];

/* palettebits is used to set the hibits of each byte in the hgr output
   raster based on color dominance by doing a pixel count of green-violet
   vs. orange-blue for each byte in the hgr raster

   orange and blue pixels take precedence over green and purple pixels */
unsigned char a2fillcolor = HBLACK, palettebits[40], paltype = 255; /* Both (default) */

/* encbuf is where I accumulate each Apple II raster sequentially before
   encoding the image (fragment or fullscreen) into a RAX file */
unsigned char encbuf[7680];


/* the following palettes are used to map BMP colors to HGR colors */

/* user defined conversion colors option through use of a BAP (palette)
   file for each BMP - applewin rendering is used as the default */
unsigned char usermap[6][4] = {
	0,0,0,      HBLACK,    /* black */
	32,192,0,   HGREEN,    /* green */
	160,0,255,  HVIOLET,   /* violet */
	240,80,0,	HORANGE,   /* orange */
	0,128,255,  HBLUE,     /* blue */
	255,255,254,HWHITE};   /* white */


/* rgb values in a bap file are in the following order
   (same order as my programming constants and just about everything
    else here in BMP2RAG-land):

 black
 green
 violet
 orange
 blue
 white

 these are comma separated values in decimal format from 0-255 for each
 r,g,b gun value and must match one of the 16 colors in the palette of the
 associated BMP file.

*/

/* cmapplewin palette array note:

   The r,g,b values for the 6 Colors in cmapplewin palette are the values
   produced by doing a Windows screen-capture of the AppleWin Emulator HGR
   Display in 32 bit color.

   This is provided just in case somebody has an editor for 16 color bmp
   files that provides custom palette support and scaling without
   producing half-tones and wants to export Apple II screen captures back
   to the Apple II as native files..

   For those of us who don't...

   In Windows XP Paint, when a screen-capture from AppleWin is pasted from
   the Windows Clipboard, and the image is then saved as a 16 color bmp,
   the colors from AppleWin will (for the most part) be remapped correctly
   when resaving back to an Apple II HGR image using my standard Windows
   XP Paint BMP palette remapping...

*/

/* clipshop rendering */
unsigned char cmappleclip[6][4]={
0x00, 0x00, 0x00, HBLACK,
255,  0x00, 128,  HVIOLET,
0x00, 192,  0x00, HGREEN,
255,  128,  0x00, HORANGE,
0x00, 0x00, 192,  HBLUE,
0xFF, 0xFF, 0xFF, HWHITE};

/* applewin rendering */
/* applewin color order */
unsigned char cmappleold[6][4] = {
  0,  0,  0,HBLACK,
255,255,254,HWHITE,
  0,128,255,HBLUE,
240, 80,  0,HORANGE,
 32,192,  0,HGREEN,
160,  0,255,HVIOLET};

unsigned char cmapplenew[6][4] = {
  0,  0,  0,HBLACK,
255,255,254,HWHITE,
 13,161,255,HBLUE,
242, 94,  0,HORANGE,
 56,203,  0,HGREEN,
199, 52,255,HVIOLET};

unsigned char cmapplefuture[6][4] = {
  0,  0,  0,HBLACK,
255,255,254,HWHITE,
 13,161,255,HBLUE,
242, 94,  0,HORANGE,
 56,203,  0,HGREEN,
199, 52,255,HVIOLET};


/* In the palettes below, the first 3 values for each of the 16 color
   entries are the BMP's r,g,b values and the last value is an ordinal 0-5
   for the Apple II's HGR colors */

/* 16 Color BMP style palette (Windows XP, Windows 7) this palette is arranged in
   Apple II HGR color order and the palettes that follow are arranged in
   IBM-PC (Windows) color order and use equivalent mapping to Apple II
   colors as the colormap array below... */
unsigned char colormap[16][4] = {
	0,0,0,0,         /* black (black) */
	128,128,128,0,   /* black (dark grey) */
	0,255,0,1,       /* green (light green) */
	0,128,0,1,       /* green (dark green) */
	128,128,0,1,     /* green (brown - dark yellow) */
	255,0,255,2,     /* violet (light magenta) */
	128,0,128,2,     /* violet (dark magenta) */
	255,0,0,3,       /* orange (light red) */
	128,0,0,3,       /* orange (dark red) */
	64,0,255,4,      /* blue (light blue) */
	0,0,128,4,       /* blue (dark blue) */
	0,255,255,4,     /* blue (light cyan) */
    0,128,128,4,     /* blue (dark cyan) */
	255,255,255,5,   /* white */
	255,255,0,5,     /* white (light yellow) */
	192,192,192,5};  /* white (light grey) */

/* 16 color ZSOFT PCPAINT PCX style palette */
unsigned char cmappcx[16][4]={
0x00, 0x00, 0x00, 0,  /* BLACK    */
0x00, 0x00, 0xAA, 4,  /* BLUE     */
0x00, 0xAA, 0x00, 1,  /* GREEN    */
0x00, 0xAA, 0xAA, 4,  /* CYAN     */
0xAA, 0x00, 0x00, 3,  /* RED      */
0xAA, 0x00, 0xAA, 2,  /* MAGENTA  */
0xAA, 0xAA, 0x00, 1,  /* BROWN    */
0xAA, 0xAA, 0xAA, 5,  /* WHITE    */
0x55, 0x55, 0x55, 0,  /* GRAY     */
0x55, 0x55, 0xFF, 4,  /* LBLUE    */
0x55, 0xFF, 0x55, 1,  /* LGREEN   */
0x55, 0xFF, 0xFF, 4,  /* LCYAN    */
0xFF, 0x55, 0x55, 3,  /* LRED     */
0xFF, 0x55, 0xFF, 2,  /* LMAGENTA */
0xFF, 0xFF, 0x55, 5,  /* YELLOW   */
0xFF, 0xFF, 0xFF, 5}; /* BWHITE   */

/* 16 Color BMP style palette (probably PBRUSH from Windows 3.1) */
unsigned char cmapbmp[16][4]={
0x00, 0x00, 0x00, 0,  /* BLACK    */
0x00, 0x00, 0xBF, 4,  /* BLUE     */
0x00, 0xBF, 0x00, 1,  /* GREEN    */
0x00, 0xBF, 0xBF, 4,  /* CYAN     */
0xBF, 0x00, 0x00, 3,  /* RED      */
0xBF, 0x00, 0xBF, 2,  /* MAGENTA  */
0xBF, 0xBF, 0x00, 1,  /* BROWN    */
0xC0, 0xC0, 0xC0, 5,  /* WHITE    */
0x80, 0x80, 0x80, 0,  /* GRAY     */
0x00, 0x00, 0xFF, 4,  /* LBLUE    */
0x00, 0xFF, 0x00, 1,  /* LGREEN   */
0x00, 0xFF, 0xFF, 4,  /* LCYAN    */
0xFF, 0x00, 0x00, 3,  /* LRED     */
0xFF, 0x00, 0xFF, 2,  /* LMAGENTA */
0xFF, 0xFF, 0x00, 5,  /* YELLOW   */
0xFF, 0xFF, 0xFF, 5}; /* BWHITE   */

/* VGA Palette probably converted to 24 bit from actual 18 bit adapter
   settings but I really can't remember */
unsigned char cmapvga[16][4]={
0x00, 0x00, 0x00, 0,  /* BLACK    */
0x00, 0x00, 0xFF, 4,  /* BLUE     */
0x00, 0xFF, 0x00, 1,  /* GREEN    */
0x00, 0xFF, 0xFF, 4,  /* CYAN     */
0xFF, 0x00, 0x00, 3,  /* RED      */
0xFF, 0x00, 0xFF, 2,  /* MAGENTA  */
0xFF, 0xFF, 0x00, 1,  /* BROWN    */
0xC0, 0xC0, 0xC0, 5,  /* WHITE    */
0x55, 0x55, 0x55, 0,  /* GRAY     */
0x55, 0x55, 0xFF, 4,  /* LBLUE    */
0x55, 0xFF, 0x55, 1,  /* LGREEN   */
0x55, 0xFF, 0xFF, 4,  /* LCYAN    */
0xFF, 0x55, 0x55, 3,  /* LRED     */
0xFF, 0x55, 0xFF, 2,  /* LMAGENTA */
0xFF, 0xFF, 0x55, 5,  /* YELLOW   */
0xFF, 0xFF, 0xFF, 5}; /* BWHITE   */


/* GetAppleIIColor attempts to match BMP colors using the palettes above
   and returns Apple II colors, and should always work with 16 color BMP
   files produced by Windows XP Paint and hopefully those other 16 color
   BMP's made elsewhere. */
unsigned char GetAppleIIColor(unsigned char red, unsigned char green, unsigned char blue)
{

    unsigned char idx;

    /* try exact match */
    if (externalmap != 0) {
		for (idx = 0; idx < 6; idx++) {
			if (red == usermap[idx][0] && green == usermap[idx][1] && blue == usermap[idx][2])return idx;
		}
	}
	/* only 6 colors in AppleWin cmap */
   	for (idx = 0; idx < 6; idx++) {
		if (red == cmappleold[idx][0] && green == cmappleold[idx][1] && blue == cmappleold[idx][2])return cmappleold[idx][3];
	}
   	for (idx = 0; idx < 6; idx++) {
		if (red == cmapplenew[idx][0] && green == cmapplenew[idx][1] && blue == cmapplenew[idx][2])return cmapplenew[idx][3];
	}
	if (applewinversion == 911) {
		for (idx = 0; idx < 6; idx++) {
			if (red == cmapplefuture[idx][0] && green == cmapplefuture[idx][1] && blue == cmapplefuture[idx][2])return cmapplefuture[idx][3];
		}
	}
    for (idx = 0; idx < 6; idx++) {
		if (red == cmappleclip[idx][0] && green == cmappleclip[idx][1] && blue == cmappleclip[idx][2])return cmappleclip[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == colormap[idx][0] && green == colormap[idx][1] && blue == colormap[idx][2])return colormap[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmapbmp[idx][0] && green == cmapbmp[idx][1] && blue == cmapbmp[idx][2])return cmapbmp[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmapvga[idx][0] && green == cmapvga[idx][1] && blue == cmapvga[idx][2])return cmapvga[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmappcx[idx][0] && green == cmappcx[idx][1] && blue == cmappcx[idx][2])return cmappcx[idx][3];
	}

	/* adjust gun values to match using EGA-like thresholds
	  this corresponds to the old PCX style palette */
	if (red < 43) red = 0;         /* 0x00 */
	else if (red < 128) red = 85;  /* 0x55 */
	else if (red < 224) red = 170; /* 0xaa */
	else red = 255;                /* 0xff */

	if (green < 43) green = 0;
	else if (green < 128) green = 85;
	else if (green < 213) green = 170;
	else green = 255;

	if (blue < 43) blue = 0;
	else if (blue < 128) blue = 85;
	else if (blue < 213) blue = 170;
	else blue = 255;

    for (idx = 0; idx < 16; idx++) {
		if (red == cmapvga[idx][0] && green == cmapvga[idx][1] && blue == cmapvga[idx][2])return cmapvga[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmappcx[idx][0] && green == cmappcx[idx][1] && blue == cmappcx[idx][2])return cmappcx[idx][3];
	}

	/* that's enough! if no match yet, garbage-in so return 255 */
    return 255;
}


unsigned char GetReverseColor(unsigned char c)
{

	   switch(c)
	   {
		   case HBLACK: if (Wreverse == 0)break;
		                c = HWHITE; break;
		   case HGREEN:if (Vreverse == 0)break;
		                c = HVIOLET; break;
		   case HVIOLET:if (Vreverse == 0)break;
		                c = HGREEN; break;
		   case HORANGE:if (Breverse == 0)break;
		                c = HBLUE; break;
		   case HBLUE:if (Breverse == 0)break;
		                c = HORANGE; break;
		   case HWHITE:if (Wreverse == 0)break;
		                c = HBLACK; break;
	   }

	   return c;


}

void ReverseColors()
{

	int idx;
	unsigned char c;

	if (Vreverse == 0 && Breverse == 0 && Wreverse == 0) return;

	for (idx = 0; idx < 282; idx ++) {
	   c = work280[idx];
	   work280[idx] = GetReverseColor(c);
	}
}




/* the following 2 routines are almost verbatim to
   those in the zsoft pcpaintbrush manual from the '80's */

/* the rax (encoded raster graphics) format is just one big chunk of pcx
   run-length encoded (RLE) data with a 2 byte header
   (width in bytes x height in rasters) tagged on the beginning of the chunk */

/* pcx encoder helper */
/* the writer for the encline function */
int encput(unsigned char byt,unsigned char cnt, FILE *pcxfile)
{
          if(cnt){
            if((cnt==1)&& (0xc0 != (0xc0 &byt))){
                if(EOF == fputc((int)byt,pcxfile))
                return(0);
                return(1);
            }
            else{
                if(EOF==fputc((int)0xc0|cnt,pcxfile))
                return(0);
                if(EOF==fputc((int)byt,pcxfile))
                return(0);
                return(2);
            }
        }
        return(0);

}

/* encode a line in pcxformat encoding */
/* encodes a raw line and writes it out to disk */
int encline(unsigned char *inbuff,int inlen, FILE *pcxfile)
{
    unsigned char this,last;
    int srcindex,i;
    int total;
    unsigned char runcount;
    total=0;
    last = *(inbuff); runcount=1;

for(srcindex=1;srcindex!=inlen;srcindex++){
    this= *(++inbuff);
    if(this==last){
        runcount++;
        if(runcount==63){
            if(!(i=encput(last,runcount,pcxfile)))
            return(0);
            total+=i;
            runcount=0;
        }
    }
    else{
        if(runcount){
            if(!(i=encput(last,runcount,pcxfile)))
            return(0);
            total+=i;
        }
        last=this;
        runcount=1;
      }
   }

   if(runcount){
    if(!(i=encput(last,runcount,pcxfile)))
    return(0);
    return(total+i);
    }
    return (total);

}


/* The following routines determine whether the hibit in each byte will be
   set or not.

   When the hibit is set in each byte, green and violet will map to orange
   and blue.

   When the hibit is not set, orange and blue will map to green and
   violet.

   I call these hibits "palette bits" throughout my stuff. Sometimes I
   will refer to palette 0 and palette 1, sometimes I will refer to the
   Green and Orange palettes when describing whether the hibit for each
   byte of the Apple II display is unset or set.

   I call green, violet, orange, and blue "colored pixels". I call the
   other two available pixels "black" and "white". Black and White ignore
   the palette bit so display in either palette and therefore do not need
   to be considered for palette selection.

   If a bmp image uses both orange-blue pixels and green-violet pixels
   horizontal regions in blocks of 7 pixels and 14 pixels must be
   considered for layout purposes.

   For best conversion results, orange-blue pixels and green-violet pixels
   should not be laid-out in the same 7 pixel or 14 pixel block and
   separation at block boundaries with black or white based on the
   conditions noted below will provide extra insurance that color integrity
   will be preserved in the Apple II output.

   In this program, I introduce the concept of base resolution.  If the
   base horizontal resolution of a BMP is 140 the BMP will be
   double-scaled so a 7 pixel layout is always ok. if base horizontal
   resolution is 280, 7 pixels will be ok too provided the last pixel on
   an "even" block that separates Green-Violet from Orange-Blue, pixel
   6,20,34, etc., is either black or white and not colored.

   When working on a bmp layout for a base horizontal resolution of 280,
   double pixel pairs (even,odd) work best. On a bmp with 140 base hr this
   doesn't matter because every pixel will be doubled and this criteria
   will always be met.

   Code for setting palette bits follows...


*/

/* called by setPaletteBits to set the palette bits for 7 bmp pixels which
   boils-down to 3.5 pixels (1 byte) on the Apple II */
void palettebits7()
{
	int idx = 0, jdx = 0, x,gcnt,ocnt;
	unsigned char *ptr;

	while (idx < 40) {
		ptr = (unsigned char *)&work280[jdx];

		/* set palette based on color dominance between green-violet and
		   orange-blue. orange-blue is the default and if the count is
		   equal the palette will be set to orange-blue */
		gcnt = ocnt = 0;
		for (x = 0; x < 7; x++) {
		   if (ptr[x] == HGREEN  || ptr[x] == HVIOLET)gcnt++;
		   if (ptr[x] == HORANGE || ptr[x] == HBLUE)ocnt++;
		}

		/* carry forward the palette for this line from the previous bmp
		   pixel group if no orange or green pixels in this byte. this
		   will correctly set the orphans on the even bytes */
		if (gcnt == 0 && ocnt == 0 && idx%2 == 1) {
		   palettebits[idx] = palettebits[idx-1];
		}
		else {
			/* set palette for 1 Apple 2 byte */
			if (gcnt > ocnt) palettebits[idx] = 0;
			else palettebits[idx] = 0x80;
		}
		jdx+=7;
		idx++;
	}
}

/* called by setPaletteBits to set the palette bits for 14 bmp pixels
   which boils-down to 7 pixels (2 bytes) on the Apple II */
void palettebits14()
{
	int idx = 0, jdx = 0, x,gcnt,ocnt;
	unsigned char *ptr;

	while (idx < 40) {
		ptr = (unsigned char *)&work280[jdx];

		/* set palette based on color dominance between green-violet and
		   orange-blue. orange-blue is the default and if the count is
		   equal the palette will be set to orange-blue */
		gcnt = ocnt = 0;
		for (x = 0; x < 14; x++) {
		   if (ptr[x] == HGREEN  || ptr[x] == HVIOLET)gcnt++;
		   if (ptr[x] == HORANGE || ptr[x] == HBLUE)ocnt++;
		}
		/* set palette for 2 Apple 2 bytes */
		if (gcnt > ocnt) palettebits[idx] = palettebits[idx+1] = 0;
		else palettebits[idx] = palettebits[idx+1] = 0x80;

		jdx+=14;
		idx+=2;
	}
}


/* called by buildAppleRaster to set the palettebits for each line */
void setPaletteBits()
{
	int x;

	if (paltype == 0 || paltype == 0x80) {
		/* single palette over-ride... 4 color output. all non-black and
		   non-white pixels will be converted to either Green-Violet or
		   Orange-Blue */
	    for (x = 0; x < 40; x++) palettebits[x] = paltype;
	}
	else {
	  /* select dominant palette based on colors of a pixel group for
		 either a 1 byte or a 2 byte group

		 this determines if a green-violet or orange-blue palette will be
		 used based-on color dominance...

		 see the code for the below above. for more information call
		 apple.

		*/
		if (x14 == 0)palettebits7();
		else palettebits14();
	}
}


void buildhgr()
{
	 int i, j;

     /* create bit pattern from pixelized values */
     /* i is even and j is odd */
	 for (i= 0, j = 1; i < 280; i+=2, j+=2) {
		buf280[i] = 0; /* assume everything is black */
		buf280[j] = 0;

        /* add the white bits - this also accounts for the half shift of
           the color pixels which applewin renders as white to represent
           aliasing of the color anomalies */

		if (work280[i] == HWHITE) buf280[i] = 1;
		if (work280[j] == HWHITE) buf280[j] = 1;
        if (doublecolors == 1) {
			/* add the violet or blue bits - the 2-bit value will be 2 */
			if (work280[i] == HBLUE || work280[i] == HVIOLET ||
			    work280[j] == HBLUE || work280[j] == HVIOLET ) {
					buf280[i] = 1;
					buf280[j] = 0;
			}

			/* add the green or orange bits - the 2-bit value will be 1 */
			if (work280[i] == HORANGE || work280[i] == HGREEN ||
			    work280[j] == HORANGE || work280[j] == HGREEN ) {
					buf280[i] = 0;
					buf280[j] = 1;
			}

		}
		else {
			/* add the violet or blue bits - the 2-bit value will be 2 */
			if (work280[i] == HBLUE || work280[i] == HVIOLET )buf280[i] = 1;
			if (work280[j] == HBLUE || work280[j] == HVIOLET )buf280[j] = 0;

			/* add the green or orange bits - the 2-bit value will be 1 */
			if (work280[i] == HORANGE || work280[i] == HGREEN )buf280[i] = 0;
			if (work280[j] == HORANGE || work280[j] == HGREEN )buf280[j] = 1;
		}

        if (doubleblack == 1) {
		  if (work280[i] == HBLACK || work280[j]==HBLACK) {
		  		buf280[i] = 0;
		  		buf280[j] = 0;
		  }
		}

	 }

}


void setAppleRaster(unsigned char *bytebuf)
{
	int x,x2,bytes;
	unsigned char bit[7], palbit, c, hgrpixel, strategy;

	/* at this point in the program our hgr color values have been moved
	into our work buffer and if our base horizontal resolution for the BMP
	is 140 rather than 280 our values have been doubled-up in duplicate
	pairs. Our palette has been set as well... */

	/* now that the palette has been set we can replace our orange values
	with green and our blue values with violet */

	for(x=0;x<282;x++) {
		if(work280[x] == HORANGE)work280[x] = HGREEN;
		else if(work280[x] == HBLUE)work280[x] = HVIOLET;
	}

    buildhgr();

    /* now pack the raster - always pack a full raster */
	x = 0;
	for(bytes=0;bytes<40;bytes++) {
		palbit = (unsigned char) palettebits[bytes];
		for(x2=0;x2<7;x2++) {
			bit[x2]=buf280[x];
			x++;
		}
		bytebuf[bytes]=(palbit|bit[6]<<6|bit[5]<<5|bit[4]<<4| bit[3]<<3|bit[2]<<2|bit[1]<<1|bit[0]);
	}
}


/* called from below... obviously not something angelic that you would
bring home to meet the parents... */
void buildAppleRaster(int y)
{
	int x, x1,x2, yoff, width;
	char *ptr, *ptr2;
	unsigned char c;

	memset(&work280[0],a2fillcolor,282);

	x2 = (int)bmp.bmiHeader.biWidth;

	/* transfer the line to our workbuffer */
	x = a2shift;
	for (x1 = 0; x1 < x2; x1++) {
	   if (x > 279) break;
	   work280[x] = buf280[x1];
	   x++;
       if (x > 279) break;
	   if (bmpformat == 96 || bmpformat == 192) {
		    /* double the line width */
	   		work280[x] = buf280[x1];
	   		x++;
	   }
	}

    ReverseColors();
    setPaletteBits();

	yoff = HB[y]-0x2000;
	ptr  = (unsigned char *) &hgrbuf[yoff];

    setAppleRaster(ptr);

    if (bmpformat == 96) {
		/* add in an second line for images
		   half-scaled in the vertical axis */
		y--;
		yoff = HB[y]-0x2000;
		ptr2  = (unsigned char *) &hgrbuf[yoff];
		memcpy(ptr2,ptr,40);
	}
}

/* unpack and translate a 16 color line to a line of Apple II color
   indices...*/
void Vga2AppleIndex(unsigned width)
{
     int i,j;
     unsigned char idx;


     for(i=0,j=0;j<width;)
     {
		/* unpack high and low nibbles of 16 color BMP scanline and remap
		   to one of 6 hgr colors */
        idx = bmpscanline[i]>>4;
        buf280[j] = bmpmapinfo[idx];
        j++;
        if (j < width) {
        	idx = bmpscanline[i]&0xf;
        	buf280[j] = bmpmapinfo[idx];
        	i++;
        	j++;
		}
     }
}


int ReadExternalMapFile(char *mapfile)
{

	unsigned char *ptr, buf[128];
	FILE *fp;
	int r,g,b,i = 0,j;

	sprintf(buf,"%s.BAP",mapfile);

    /* user defined color option */
    if((fp=fopen(buf,"r"))!=NULL) {
	  while (i<6) {
		if (NULL == fgets(buf,80,fp)) break;
		if (buf[0] < 48 || buf[0] > 57) continue;
		ptr = (char *)&buf[0];
		r = atoi(ptr);
		if (r < 0 || r > 255) continue;
		g = -1; b = -1;
		for (j=0;buf[j]!=0;j++) {
			if (buf[j] == ',') {
				ptr = (char *)&buf[j+1];
				if (g == -1) g = atoi(ptr);
				else b = atoi(ptr);
				if (g < 0||g > 255) break;

			}
			if (b!=-1) {
				if (b < 0||b > 255) break;
				usermap[i][0] = (unsigned char)r;
				usermap[i][1] = (unsigned char)g;
				usermap[i][2] = (unsigned char)b;
				i++;
				break;
			}
		}
	  }
	  fclose(fp);
	}
	return i;

}


void fillhgrbuf()
{
	int i, j, xoff;
	unsigned char palettebit = 0,c1,c2, bits[7], *ptr, fillcolor;

    memset(&hgrbuf[0],0,8192); /* clear write buffer */

    fillcolor = GetReverseColor(a2fillcolor);
	if (fillcolor < 1 || fillcolor > 5) {
		return;
	}

	switch(fillcolor)
	{
		case HWHITE: c1 = c2 = 1; break;

		case HORANGE: palettebit = 0x80;
		case HGREEN:  c1 = 0; c2 = 1; break;

		case HBLUE:   palettebit = 0x80;
		case HVIOLET: c1 = 1; c2 = 0;
	}

	if (paltype == 0 || paltype == 0x80) {
		/* single palette over-ride... 4 color output. all non-black and
		   non-white pixels will be converted to either Green-Violet or
		   Orange-Blue */
		palettebit = paltype;
	}

	for (i= 0, j = 1; i < 280; i+=2, j+=2) {
		buf280[i] = c1;
		buf280[j] = c2;
	}

	for (i = 0, xoff=0; i < 40; i++) {

		for (j = 0; j < 7; j++) {
			bits[j] = buf280[xoff]; xoff++;
		}
		c1 = (palettebit | bits[6]<<6|bits[5]<<5|bits[4]<<4|
			bits[3]<<3|bits[2]<<2|bits[1]<<1|bits[0]);

		palettebits[i] = c1;
	}

	for (i = 0; i < 192; i++) {
	    xoff = HB[i]-0x2000;
	    ptr  = (unsigned char *) &hgrbuf[xoff];
		for (j = 0; j < 40; j++) ptr[j] = palettebits[j];

	}

}


/* read and convert a 16 color BMP file in one of our 3 supported
resolution formats:

base resolution
	- 140 x 96  (commandline option 96 - "half-scale")
	- 140 x 192 (commandline option 192 - "half-width")
	- 280 x 192 (default)

Apple II image fragments are produced either using the default or the base
resolution specified on the commandline. */
int BMPRead(char *basename)
{

    FILE *fp;
    int i,j,cnt,r,g,b;
    unsigned int wHres,wVres,packet;
    char bmpfile[128];
    unsigned char c, *ptr;

	sprintf(bmpfile,"%s.BMP",basename);
    if((fp=fopen(bmpfile,"rb"))==NULL)return -1;

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmp.bmiHeader.biSize,
                 sizeof(VGABITMAPINFO),1,fp);


    /* if the BMP is wider than 140 then silently and
       automatically adjust to 280 x 192 base resolution and
       if the height of a bmp with a base resolution of 140 x 96
       is greater than 96, do the same.
       */
    if (bmp.bmiHeader.biWidth > 140 || (bmpformat == 96 && bmp.bmiHeader.biHeight > 96)) {
		bmpformat = 280;
		maxwidth = 280;
		maxheight = 192;
	}


    /* the 256 color and 24-bit bmps are tested first... by the time it gets to here
       if it fails again then the user is out of options so post an error on
       our way out... */
    if(BitMapFileHeader.bfType[0] != 'B' || BitMapFileHeader.bfType[1] != 'M' ||
	   bmp.bmiHeader.biCompression!=BI_RGB ||
       bmp.bmiHeader.biWidth > maxwidth ||
       bmp.bmiHeader.biHeight> maxheight ||
       bmp.bmiHeader.biPlanes!=1 ||
       bmp.bmiHeader.biBitCount!=4
       )
       {
		   fclose(fp);
           if (BitMapFileHeader.bfType[0] != 'B' || BitMapFileHeader.bfType[1] != 'M')
           		puts("Input file is not a BMP file!");
		   else if (bmp.bmiHeader.biCompression!=BI_RGB)
		   		puts("Run-length encoded BMPs are not supported!");
		   else if (bmp.bmiHeader.biPlanes!=1 || bmp.bmiHeader.biBitCount!=4)
		   		puts("BMP must be in 16 color packed-pixel format!\nOr a single plane 256 color or 24-bit BMP!");
		   else if (bmp.bmiHeader.biWidth > maxwidth || bmp.bmiHeader.biHeight> maxheight)
		        puts("BMP is too large and will not be processed!");

           return -1;
       }


    for(i=0,cnt=0;i<16;i++) {
	/* check for user defined colors from a bap (colormap) file a colormap
	   file gives the user the option to use bmp editors other than
	   Windows Paint which may provide a different palette.

	   there is no point in guessing what bmp color to use to remap to
	   apple colors when the editor is different but having said that, I
	   still try when no palette file is present.

	   */
		if (externalmap != 0) {
			for (j=0;j<externalmap;j++) {
			   if (bmp.bmiColors[i].rgbRed == usermap[j][0] &&
				   bmp.bmiColors[i].rgbGreen == usermap[j][1] &&
				   bmp.bmiColors[i].rgbBlue == usermap[j][2]) {
					   bmpmapinfo[i] = (unsigned char)j;
					   cnt++;
					   break;
				   }
			}
		    continue;
		}
		/* if no user color mapping use our built-in map... */
        bmpmapinfo[i] = c = (unsigned char) GetAppleIIColor(bmp.bmiColors[i].rgbRed,
                                            				bmp.bmiColors[i].rgbGreen,
                                        					bmp.bmiColors[i].rgbBlue);
        if (c!=255)cnt++;

        if (BUGS) {
			if(i==0) printf("unsigned char %smap[16][4] = {\n",basename);
			printf("%3d, %3d, %3d, %d", bmp.bmiColors[i].rgbRed,
									 	bmp.bmiColors[i].rgbGreen,
									 	bmp.bmiColors[i].rgbBlue,c);
			if(i==15)printf("};\n");
			else printf(",\n");
	    }

    }

    if (cnt == 0) {
		fclose(fp);
		puts("Error! Unable to build a colormap...");
		return -1;

	}
/* GIGO (garbage-in, garbage-out) set the invalid colors in the colormap
   to black and hopefully they have not been used */
    for (i = 0; i < 16; i++) {
		if (bmpmapinfo[i] == 255)bmpmapinfo[i] = 0;
	}


	/* it is fine to use a colormap of less than 6 colors if the bmp file
	   has fewer than 6 colors used but warn them anyway... */
	if (cnt < 16) {
	    printf("Warning! Incomplete color map...\n%d colors will be set to black if used!\n",16-cnt);
	}
	puts("Color map built successfully!");

    wHres = (unsigned int )bmp.bmiHeader.biWidth;
    packet = wHres/2;
    while(packet%4!=0)packet++;
    wVres = (unsigned int )bmp.bmiHeader.biHeight;


/* before we build our Apple II image (or fragment) we set-up the width in
   bytes x height in rasters for our Apple II HGR output image because we
   are here now doing this sort of stuff */
    if (bmpformat == 280)
    	ragwidth = (unsigned int )bmp.bmiHeader.biWidth;
    else
    	ragwidth = (unsigned int )bmp.bmiHeader.biWidth * 2;

    ragwidth += a2shift;
    while (ragwidth > 280) {
	   ragwidth--;
	   a2shift--;
	}
    ragpixels = ragwidth;
    while (ragwidth%7 !=0)ragwidth++; /* multiples of 7 pixels */
    /* if we have an orphan pixel hanging at the edge of an even byte
       increase the width to the next 7 pixels */
    if (ragpixels == ragwidth && (ragwidth%14)!=0) {
		ragwidth+=7;
	}
	ragwidth = (ragwidth/7);
	if (ragwidth > 40) ragwidth = 40;
	if (ragpixels > 280) ragpixels = 280;

	if (bmpformat == 96)
		ragheight = (unsigned int )bmp.bmiHeader.biHeight*2;
	else
		ragheight = (unsigned int )bmp.bmiHeader.biHeight;

    fillhgrbuf();
    printf("RagWidth = %d bytes, RagHeight = %d scanlines.\n",ragwidth,ragheight);

    memset(&bmpscanline[0],0,140);

    for(i=wVres;i>0;i--)
    {
	  fread((char *)&bmpscanline[0],1,packet,fp);

      /* conversion to Apple II HGR image format
         raster by raster */

      /* unpack the line */
      Vga2AppleIndex(wHres);

      /* now do-it */
      switch(bmpformat) {
	  	case 96:
	  		buildAppleRaster((i*2)-1);break;
	  	case 192:
	  	    buildAppleRaster(i-1);break;
	  	case 280:
	  	    buildAppleRaster(i-1);
	  }
    }
    fclose(fp);

return 0;
}


/* rgb triples from applewin screen grab using windows clipboard
   or svga indices from applewin screen save */
unsigned char AppleWinLine[840];

/* converting from applewin bmp's back to hgr should be accurate
   since all we are doing is reversing applewin's  rendering */
void UnpackAppleWinLine()
{
     int i,j,k, green, orange;
     unsigned char r, g, b, p;


     k = a2shift;
     for (i = 0; i < 282; i++) work280[i] = a2fillcolor;

     /* convert rgb triples or svga indices to HGR colors */
     for(i=0,j=0;i<bmi.biWidth;i++) {
		 /* 24 bit */
		 if (bmi.biBitCount == 24) {
		 	b =  AppleWinLine[j]; j++;
		 	g =  AppleWinLine[j]; j++;
		 	r =  AppleWinLine[j]; j++;
		 }
		 /* 8 bit */
		 if (bmi.biBitCount == 8) {
			 p = AppleWinLine[j]; j++;
			 r = sbmp[p].rgbRed;
             g = sbmp[p].rgbGreen;
             b = sbmp[p].rgbBlue;
		 }
		 work280[k] = GetAppleIIColor(r,g,b);
		 k++;
		 if (k > 279) break;
	}

    ReverseColors();

    buildhgr();

     /* set the palette based on groups of seven pixels */
	if (paltype == 0 || paltype == 0x80) {
		/* single palette over-ride... 4 color output. all non-black and
		   non-white pixels will be converted to either Green-Violet or
		   Orange-Blue */
		for (i = 0; i < 40; i++) palettebits[i] = paltype;
	}
	else {
		 p = 0x80;
		 for (i = 0, k=0; i < 40; i++ ) {
			 orange = 0;
			 green = 0;
			 for (j = 0; j < 7; j++) {
				 if (work280[k] == HORANGE ||  work280[k] == HBLUE) orange++;
				 if (work280[k] == HGREEN ||  work280[k] == HVIOLET) green++;
				 k++;
			 }
			 if (green > orange) p = 0;
			 if (orange > green) p = 0x80;

			 palettebits[i] = p;
		 }
	 }

}


/* encodes AppleWin apple II hgr scanline into buffer */
int applebits(int y)
{
	    int xoff,idx,jdx;
	    unsigned char *ptr, bits[7], x1, palettebit;

	    xoff = HB[y]-0x2000;
	    ptr  = (unsigned char *) &hgrbuf[xoff];

        xoff = 0;
        for (idx = 0; idx < 40; idx++) {

            for (jdx = 0; jdx < 7; jdx++) {
				bits[jdx] = buf280[xoff]; xoff++;
			}
			palettebit = palettebits[idx];

			x1 = (palettebit | bits[6]<<6|bits[5]<<5|bits[4]<<4|
                        bits[3]<<3|bits[2]<<2|bits[1]<<1|bits[0]);

            ptr[idx] = x1;
		}
		return SUCCESS;

}





int AppleWinToHGR(unsigned char *basename)
{

 	FILE *fp;
    int packet = INVALID, y,i,j;
    char bmpfile[128];

	sprintf(bmpfile,"%s.BMP",basename);

    if((fp=fopen(bmpfile,"rb"))==NULL) {
		return INVALID;
	}

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);

    if (bmi.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmi.biPlanes==1) {

		if (bmi.biBitCount == 24 && bmi.biWidth < 281 && bmi.biHeight < 193) {
			packet = bmi.biWidth * 3;
		}
		if (bmi.biBitCount == 8 && bmi.biWidth < 281 && bmi.biHeight < 193) {
			packet =  bmi.biWidth;
			fread((char *)&sbmp[0].rgbBlue, sizeof(RGBQUAD)*256,1,fp);

			/* check applewin version - some of this for future versions */
			for (i = 16,j=0; j < 6; i++,j++) {
				 applewinversion = 0;
				 if (sbmp[i].rgbRed != cmapplenew[j][0]) break;
				 if (sbmp[i].rgbGreen != cmapplenew[j][1]) break;
				 if (sbmp[i].rgbBlue != cmapplenew[j][2]) break;
				 applewinversion = 120;
			}
			if (applewinversion == 0) {
				for (i = 16,j=0; j < 6; i++,j++) {
					 applewinversion = 0;
					 if (sbmp[i].rgbRed != cmappleold[j][0]) break;
					 if (sbmp[i].rgbGreen != cmappleold[j][1]) break;
					 if (sbmp[i].rgbBlue != cmappleold[j][2]) break;
					 applewinversion = 116;
				}
			}
			if (applewindex == 1 && applewinversion == 0) {
				applewinversion = 911;
				for (i = 16,j=0; j < 6; i++,j++) {
					cmapplefuture[j][0] = sbmp[i].rgbRed;
					cmapplefuture[j][1] = sbmp[i].rgbGreen;
					cmapplefuture[j][2] = sbmp[i].rgbBlue;
				}
			}
		}
	}

    if(packet == INVALID){ fclose(fp);return INVALID;}

    while(packet%4!=0)packet++;


/* before we build our Apple II image (or fragment) we set-up the width in
   bytes x height in rasters for our Apple II HGR output image because we
   are here now doing this sort of stuff */
    ragwidth = (unsigned int )bmi.biWidth + a2shift;

    while (ragwidth > 280) {
		ragwidth--;
		a2shift--;
	}

    ragpixels = ragwidth;
    while (ragwidth%7 !=0)ragwidth++; /* multiples of 7 pixels */
    /* if we have an orphan pixel hanging at the edge of an even byte
       increase the width to the next 7 pixels */
    if (ragpixels == ragwidth && (ragwidth%14)!=0) {
		ragwidth+=7;
	}
	ragwidth = (ragwidth/7);
	if (ragwidth > 40) ragwidth = 40;
	if (ragpixels > 280) ragpixels = 280;

	ragheight = (unsigned int )bmi.biHeight;

    fillhgrbuf();

    printf("RagWidth = %d bytes, RagHeight = %d scanlines.\n",ragwidth,ragheight);
    for(y=ragheight;y>0;y--)
    {
	  fread((char *)&AppleWinLine[0],1,packet,fp);
	  UnpackAppleWinLine();
	  applebits(y-1); // encode 7 bit
    }
    fclose(fp);

return SUCCESS;

}

/* makerag is called from below... some weak attempt at information hiding
   or decomposition and abstracting maybe but I couldn't even be bothered
   to camel-case the function name... it's cute tho', eh?
   */
int makerag(FILE *fp)
{
	/* make an Rasterized Apple II Graphic (RAG) */
	/* or is it a FRAGMENT without the F and the MENT ? */
	/* the latter - I have confined my naming to MS-DOS 8.3 characters. As
	   long as DOSBox and other VM's are around this stuff will work
	   forever, virtually...

	   But don't ask me again or I'll deny the whole thing and tell you
	   the Rasterized Apple II Graphics story with the equivalent RAX and
	   RAT thrown-in to make all this nonsense more believable. Hey it
	   worked for the big-guys didn't it... still is too.

	   now, what about this code? if anyone was paying attention up to
	   now they would notice that I'm peeling this out of an
	   Apple II screen stuck right in this program a raster at a time.

	   What could be so simple? No Canvas Metaphors, Device Contexts,
	   GWorlds or XLaborOfTheMonth to draw-upon... just raw memory.

	   */

    int x, y, xoff, cnt = 0;
    unsigned char *ptr;

    for (y = 0; y < ragheight; y++) {
	    xoff = HB[y] - 0x2000;
    	ptr  = (unsigned char *) &hgrbuf[xoff];
		fwrite(ptr,1,ragwidth,fp);
		/* save rasters in pcxbuf for later
		   we'll encode it a one big chunk with 2 bytes
		   as a header */
		for (x = 0; x < ragwidth; x++) {
           encbuf[cnt] = ptr[x];
           cnt++;
		}

	}
    return cnt;
}


/*
  Apple II Embedded Graphics Text File Output
*/
void makerat(FILE *fp, FILE *fprax, char *basename, unsigned cnt)
{

		unsigned int xsize;
		int x, y, idx;

		/* the c array contains width in bytes, height in rasters
		followed by data */
		fprintf(fp,"/* %s.rat Embedded Graphics Image Created by\n"
					"   Bmp2Rag(C) Copyright Bill Buckels 2013. All Rights Reserved. */\n\n",basename);

		fprintf(fp,"/* suitable for embedding in Aztec C Apple II Programs */\n");
		fprintf(fp,"/* width in bytes, height in rasters, data... */\n\n");

		fprintf(fp,"#define RAW 1\n\n#ifdef RAW\n");
		fprintf(fp,"/* raw array of Apple II RAG image */\n");
		fprintf(fp,"unsigned char %s[%d] = {\n%3d,%3d,",basename,cnt+2,ragwidth,ragheight);
		fflush(fp);

		idx = 2;
		for (y = 0; y < cnt; y++) {
		  x=encbuf[y];
		  fprintf(fp,"%3d",x);
		  if(y == (cnt-1))fprintf(fp,"};");
		  else fputc(',',fp);
		  idx++;
		  if (idx==11) {
			  fputc('\n',fp);
			  fflush(fp);
			  idx=0;
		  }
		}

		xsize = (unsigned int)filelength(fileno(fprax));
		rewind(fprax);
		fprintf(fp,"\n#else\n");
		fprintf(fp,"/* encoded array of Apple II RAG image */\n");
		fprintf(fp,"unsigned char %s[%u] = {\n",basename,xsize);

		idx = 0;
		x= fgetc(fprax);
		fprintf(fp,"%3d",x);
		do {
			x = fgetc(fprax);
			if (x == EOF) {
				fprintf(fp,"};\n");
				break;
			}
			fputc(',',fp);
			idx++;
			if (idx == 11) {
				fputc('\n',fp);
				fflush(fp);
				idx = 0;
			}
			fprintf(fp,"%3d",x);

		}while (x!=EOF);

		fprintf(fp,"#endif\n");

}


/* save apple II hgr files... */
int savehrfiles(unsigned char *basename)
{
	FILE *fp, *fprax;
	unsigned char binfile[128], ragfile[128], raxfile[128], ratfile[128];
	unsigned int cnt;


    /* if our data segment gets clobbered chances are pretty good that
       these globals are going to get wiped by the time we get to
       here... this program has a debug switch... just type
       debug with all that other commandline nonsense... */
    if (BUGS) {
   		printf("ragwidth = %d, ragheight = %d\n",ragwidth,ragheight);
	}

    /* I am sticking with my old naming conventions...
       A BOTtom, a TOP, and a RAG...
       read all about this elsewhere... my code is riddled with it */
    if (ragwidth == 40 && ragheight == 192) {
		sprintf(ragfile,"%s.BOT",basename);
		binoutput = 1;
	}
	else if (ragwidth == 40 && ragheight == 160) sprintf(ragfile,"%s.TOP",basename);
	else sprintf(ragfile,"%s.RAG",basename);

    /* for the BASIC folks, a binary crumb */
    if (binoutput == 1) {
		sprintf(binfile,"%s.BIN",basename);
		fp = fopen(binfile,"wb");
		if (NULL == fp)return INVALID;
		fwrite(hgrbuf,1,8192,fp);
		fclose(fp);
	}

	fp = fopen(ragfile,"wb");
	if (NULL == fp) {
		if (binoutput == 1)remove(binfile);
		return INVALID;
	}

    sprintf(raxfile,"%s.RAX",basename);
    fprax = fopen(raxfile,"wb");
    if (NULL == fprax) {
		fclose(fp);
		remove(ragfile);
		if (binoutput == 1)remove(binfile);
	}

    /* width in bytes x height in rasters */
    fputc((char)ragwidth,fp);
    fputc((char)ragheight,fp);
    fputc((char)ragwidth,fprax);
    fputc((char)ragheight,fprax);

    /* write the raw raster file */
	cnt = makerag(fp);
	fclose(fp);
	/* write the encoded raster file */
	encline((unsigned char *)&encbuf[0],cnt,fprax);
	fclose(fprax);

    if (textoutput != 0) {
    	/* write the text file */
		sprintf(ratfile,"%s.RAT",basename);
		fp = fopen(ratfile,"w");
		if (NULL != fp) {
			fprax = fopen(raxfile,"rb");
			if (NULL != fprax) {
				makerat(fp, fprax,basename,cnt);
				fclose(fprax);
				fclose(fp);
			}
			else {
				fclose(fp);
				remove(ratfile);
			}
		}
	}

	return SUCCESS;
}


void SetArgs(char *argptr)
{

	int argval,idx,len,b=0,g=0,v=0,w=0;
	unsigned char c,d,e, *ptr;

	/* check for numeric commandline args first */
	argval = atoi(argptr);

	switch(argval) {
		case 280:
			bmpformat = 280;
			maxwidth = 280;
			maxheight = 192;
			break;
		case 192:
			bmpformat = 192;
			maxwidth = 140;
			maxheight = 192;
			break;
		case 96:
			bmpformat = 96;
			maxwidth = 140;
			maxheight = 96;
			break;
		default:


			/* check for alpha-numeric commandline args */
			len = strlen(argptr);
            if (len == 1) {
			    c=toupper(argptr[0]);
			    switch(c)
			    {
				   case 'O': paltype = 0x80; break;
				   case 'B': Breverse = 1;   break;
				   case 'G': paltype = 0;    break;
				   case 'V': Vreverse = 1;   break;
				   case 'W': Wreverse = 1;   break;
				   case 'I': Wreverse = 1; Breverse = 1; Vreverse = 1; break;


				}
				break;
			}



			if (len < 2 || len > 5) break;

			for (idx = 0; idx < len; idx++) {
			   /* convert to uppercase */
			   c=toupper(argptr[idx]);
			   argptr[idx] = c;
			}

			c = argptr[0];
			if (c == 'S') {
				/* shift right by ??? pixels - offset option */
				ptr = (unsigned char *)&argptr[1];
				a2shift = atoi(ptr);
				if (a2shift < 0 || a2shift > 139) a2shift = 0;
				break;
			}
			if (c == 'F') {
				/* fillcolor  - HCOLOR number  */
				ptr = (unsigned char *)&argptr[1];
				idx = atoi(ptr);
				if (idx > 0 && idx < 6) a2fillcolor = idx;
				break;
			}

			if (c == 'D') {
				d = argptr[1];
				if (d == 'C') doublecolors = 1;
				if (d == 'B') doubleblack  = 1;
			}

			switch(len) {

				case 3:
					c = argptr[0];
					d = argptr[1];
					e = argptr[2];

					/* output programming files */
					if (c == 'R' && d == 'A' && e == 'T')textoutput = 1;
					if (c == 'T' && d == 'X' && e == 'T')textoutput = 1;
					/* output bin file for fragments */
					if (c == 'B' && d == 'I' && e == 'N')binoutput = 1;
					if (c == 'B' && d == 'A' && e == 'S')binoutput = 1;
					/* expand palette bit setting from 1 to 2 bytes
					   7 bmp pixels to 14 bmp pixels */
					if (c == 'P' && d == '1' && e == '4')x14 = 1;
					break;

				case 5:
					/* debug option */
					c = argptr[0];
					d = argptr[1];
					e = argptr[2];
                    if (c == 'D' && d == 'E' && e == 'B' && argptr[3] == 'U' && argptr[4] == 'G') {
						BUGS = 1;
						break;
					}

		}
	}

}

void main(int argc, char **argv)
{
	int status,args, len;
	unsigned char fname[128],sname[128],*wordptr,scratchbuf[128];

	puts(szTextTitle);

	if(argc == 1) {
		for (;;) {
			fname[0] = ASCIIZ;
    		printf("Enter BMP Name (H for Help - Blank to Exit): ");
			gets(fname);
			if (fname[0] == ASCIIZ) {
				puts("Have a Nice Dos!");
				exit(1);
			}
			len = strlen(fname);
			/* 1.BMP = 5 characters */
			if (len > 4)break;
			puts(szTextTitle);
			puts(szUsage);
		}
	}
	else {

		strcpy(fname,argv[1]);
		len = strlen(fname);

		/* 1.BMP = 5 characters
		   garbage-in so display help */
		if (len < 5) {
			puts(szUsage);
			exit(1);
		}

		for (args = 2; args < argc; args ++) {

			if (argv[args][0] == '-')
				strcpy(scratchbuf,(char *)&argv[args][1]);
			else
				strcpy(scratchbuf,(char *)&argv[args][0]);

			SetArgs((char *)&scratchbuf[0]);
		}

	}

	strcpy(sname, fname);
	wordptr = strtok(sname, ".");

	if (BUGS) {
	     printf("Some Options: Fill=%d Shift=%d PalType=%d\n",a2fillcolor,a2shift,paltype);

	}

    externalmap = ReadExternalMapFile(sname);
	status = AppleWinToHgr(sname);
	if (status != 0)
	   status = BMPRead(sname);

	if (status!=0) {
		printf("%s.BMP is an Unsupported Format or cannot be opened.\n", sname);
		exit(1);
	}

	status = savehrfiles(sname);

	if (status == SUCCESS)printf("%s Successfully Converted!\n",fname);
	else printf("Error converting %s\n",fname);

	puts("Have a Nice Dos!");
	/* that's enough ! */
	exit(0);

}
